N_PARTICLES equ 143
DIST_MIN equ 8
GRAVITY equ 96

org 100h
  mov al,13h
  int 10h

  add dh,al
  mov gs,dx  ; z-buffer
  add dh,al
  mov fs,dx  ; back buffer
  push 0xA000
  pop es

PAL:
  mov dx,3C8h
  mov ax,cx
  out dx,al
  inc dx
  out dx,al
  shr al,1
  out dx,al
  shr al,1
  out dx,al
  loop PAL

M:
; clear z-buffer, copy backbuffer
  xor di,di
  xor si,si
CLR:
  mov word[gs:di],0
  fs movsw
  mov word[fs:di-2],0
  or di,di
  jnz CLR
  
  mov sp,-8

;  mov bx,1     ; pseudorandom seed
  in al,40h
  or ax,0xFF01
  mov bx,ax

; generate patricles: struct Particle { u16 X,Y,Z,Active; }
;;G:
;;;  imul bx,85
;;;  mov ax,bx
;;;  sar ax,1
;;;  push ax     ; random(-0.25..0.25)
;;
;;  push sp     ; [offset 6] .Active: yes
;;
;;  imul bx,85
;;  push bx     ; [offset 4] .Z: random(-0.5..0.5)
;;
;;  imul bx,85
;;  mov ax,bx
;;  or ah,80h
;;  push ax     ; [offset 2] .Y: random(-0.5..0)
;;
;;  imul bx,85
;;  push bx     ; [offset 0] .X: random(-0.5..0.5)
; 
G:
  imul bx,85
  mov ax,bx
  or ah,80h
  push ax     ; [offset 4] .Z: random(-0.5..0.5)

  or sp,sp
  js G

; loop until there's one particle left (on average)
  mov cx,255
L:push cx

; for each particle
  mov si,N_PARTICLES*8 + 0x8000
I:
  cmp word[si+6],DIST_MIN ; active?
  jb EI

 ; find the nearest neighbor
  mov ax,65535  ; ax = best distance; bx = index (uninitialized)
  mov bx,si
  mov di,N_PARTICLES*8 + 0x8000
J:
  cmp di,si ; i!=j?
  je EJ
  cmp word[di+6],DIST_MIN ; active?
  jb EJ

  ; compute squared distance dX^2 + dY^2 + dZ^2
  push di
  push si
  push ax
  xor bp,bp  ; sum
  mov cl,3
DJ:
  lodsw
  sub ax,[di]
  
  imul ax    ; dx = dX^2 | dY^2 | dZ^2 (14-bit)
  add bp,dx  ; sum <= 48k
  scasw  ; di+=2
  loop DJ
  pop ax
  pop si
  pop di

  ; new best?
  cmp ax,bp
  jb EJ
  xchg ax,bp
  mov bx,di
EJ:
  sub di,8
  js J

 ; bx = nearest_neighbor
 ; kill this particle if the neighbor is very close
  mov [si+6],ax

 ; move this particle towards the nearest neighbor
  mov cl,3
MV:
  lodsw
  sub ax,[bx]

ISAR: ; bound the delta
  sar ax,1
  imul dx,ax,-128
  jo ISAR

  imul bp,85
  push bp
  sar bp,9
  add ax,bp
  pop bp
  
  sub [si-2],ax
  inc bx
  inc bx
  loop MV
  sub si,6

  add word[si+2],GRAVITY  ; .Y += gravity

 ; draw this particle
 ; al=color ah=depth di=size bx=X bp=Y: [0 0] = center of screen

;  mov dx,[si]  ; .X / 16
;  sar dx,4
;  idiv word[si+6]
;  xchg bx,ax ; X = .X / .Z

;  mov dx,[si+2]  ; .Y / 16
;  sar dx,4
;  idiv word[si+6]
;  xchg bp,ax ; Y = .Y / .Z

  pop cx
  push cx
  mov ax,[si+4]  ; Z -> 0..32
  shr ax,10
  call DRAW

EI:
  sub si,8
  js I
  pop cx
  loop L

  in al,60h
  cmp al,1
  jne M
  int 20h

DRAW:  ; cx = iter
  pusha
 
;  movsx bx,byte[si+1]  ; X
;  movsx bp,byte[si+3]  ; Y
  mov bx,[si]  ; X
  sar bx,7
  mov bp,[si+2]  ; Y
  sar bp,7

  mov di,17
  shr cx,4
  sub di,cx      ; size ~ iter

; al=color ah=depth di=size bx=X bp=Y: [0 0] = center of screen

  imul dx,di,2
PY:
  imul cx,di,2
PX:
  pusha
  sub dx,di
  sub cx,di
  add bp,dx
  add bx,cx

  imul dx,dx  ; outside of the circle?
  imul cx,cx
  add cx,dx
  add cx,di
  imul di,di
  cmp cx,di
  jg PE

  imul di,bx,205  ; out of screen? (column 0 is unfilled)
  jo PE
  imul di,bp,320
  jo PE
  lea di,[di+bx+100*320+160]
  cmp al,[gs:di] ; z-buffer test
  jb PE
  mov [fs:di],al
  mov [gs:di],al
PE:
  popa

  loop PX
  dec dx
  jnz PY

  popa
  ret

